(*****************************************************************************)
(*                                                                           *)
(*                   DIR.INC  -  Directory in Programmen                     *)
(*                                                                           *)
(*                          ---  Vers  2.0  ---                              *)
(*                                                                           *)
(*****************************************************************************)
(* Bei Compilern, die ein mehrfaches Vorkommen von Deklarationen nicht er-
   lauben, muessen die Konstanten- und Typenvereinbarungen in den Ver-
   einbarungsteil des Hauptprogramms uebernommen werden.                     *)

Const MaxEntries = 64;           (* Maximal moegliche Eintraege im Directory *)


Type  DirStrg    = String [14];               (* Dateispez.: max. 14 Zeichen *)
      DirEntry   = Record
                      FileName :DirStrg;
                      RecSize  :Integer;
                      kSize    :Integer;
                   End;
      DirArray   = Array [1..MaxEntries] of DirEntry;
      DirOptions = (Sort,Size,CON,LST);                (* moegliche Optionen *)
      DirOptLst  = Set of DirOptions;

(*---------------------------------------------------------------------------*)

Procedure Directory (    DirUser    :Integer;
                         DirMask    :DirStrg;
                     Var DirList    :DirArray;
                     Var NoEntries  :Integer;
                         Options    :DirOptLst);

Type FCB = Record                                 (* CP/M File Control Block *)
             User      :Byte;
             FileName  :Array [1..11] of Char;
             Data      :Array [1..20] of Byte;
           End;
     DMATyp = Array [0..3] of FCB;


Var  DMA   :DMATyp;                                       (* Disk-E/A-Puffer *)
     Mask  :FCB;                           (* FCB fuer Directory-Operationen *)
     i,j,k :Integer;
     Drive,
     p     :Byte;

   (*------------------------------------------------------------------------*)
   (* CP/M neuen E/A-Puffer fuer Diskettenoperationen mitteilen.             *)

   Procedure SetDMA (Var DMAPuffer :DMATyp);

   Begin
     BDOS (26, Addr (DMAPuffer));
   End;

   (*------------------------------------------------------------------------*)
   (* CP/M ersten, zur Dateispez. passenden Directory-Eintrag suchen lassen:
      Funktionswert = 0..3 : Index in DMA-Puffer des gefundenen Eintrags,
                      255  : nichts passendes gefunden!                      *)

   Function SearchFirst (Var Mask :FCB) :Integer;

   Begin
     SearchFirst := BDOS (17, Addr (Mask));
   END;

   (*------------------------------------------------------------------------*)
   (* CP/M naechsten, passenden Eintrag suchen lassen: s. SearchFirst ()
      Achtung: zwischen SearchFirst und SearchNext duerfen keine (!) anderen
      Disketten-Operationen durchgefuehrt werden !                           *)

   Function SearchNext :Integer;

   Begin
     SearchNext := BDOS (18);
   End;

   (*------------------------------------------------------------------------*)
   (* FCB erstellen: CP/M-Dateispez. in FCB-gerechte Struktur umsetzen.
      Neuere BS wie CP/M-Plus stellen dazu ebenfalls eine BDOS-Funktion zu
      Verfuegung (CP/M-Plus: Funktion 152: Parse Filename.                   *)

   Procedure ParseFname (Var Mask :FCB; Var Drive :Byte; DirMask :DirStrg);

   VAR i, j :Integer;

   Begin
     Mask.User := 0;                       (* angemeldetes Laufwerk annehmen *)
     For i:=1 to 11 do
       Mask.FileName[i]:='?';
     For i:=1 to 20 do
       Mask.Data[i]:=0;
     For i:=1 to Length (DirMask) do             (* in FCB nur Grossbuchst.  *)
       DirMask[i] := UpCase (DirMask[i]);
     j := 1;
     i := Pos (':', DirMask);               (* anderes Laufwerk gewuenscht ? *)
     If i = 2 then
       Case DirMask[1] of
          'A'..'P': Mask.User := Ord (DirMask[1]) - Ord ('A') + 1;
       End;
     Drive := Mask.User;
     i := Succ (i);
     If i<=Length (DirMask) THEN
       With Mask do                     (* Umsetzen der Wildcards '*' in '?' *)
         Repeat
           Case DirMask[i] of
             '*': If j<8 then j:=9 else j:=12;
             '.': Begin
                    For k:=j to 8 do
                       FileName[k]:=' ';
                    j := 9
                   End
             else
               Begin
                 FileName[j] := DirMask[i];
                 j := succ (j);
               End
           End;
           i := succ (i)
         until j>11;
   End;

   (*------------------------------------------------------------------------*)
   (* Dateigroesse ermitteln: Wie zur Erstellung eines FCB gibt es in neueren
      BS dazu ebenfalls eine BDOS-Funktion. Z.B. CP/M-Plus:
      Funktion 35: Compute File Size.
      Die hier verwendete Prozedur benutzt die Turbo-Pascal-Funktion
      FileSize(f).                                                           *)

   Procedure ComputeFileSize (Var DirList   :DirArray;
                                  Drive     :Byte;
                                  NoEntries :Integer);

   Var i    :Integer;
       Name :DirStrg;
       f    :File;

   Begin
     For i:=1 to NoEntries do
       With DirList[i] do
         Begin
           Name := FileName;
                      (* Leerzeichen aus Namen entfernen und Punkt einfuegen *)
           Name := Copy (Name,1,Pos(' ',Name)-1) + '.' + Copy (Name,10,3);
                                   (* Laufwerkspez. einfuegen, z.B. 'M:....' *)
           If Drive > 0 then
             Insert (Chr (Drive - 1 + Ord('A')) + ':', Name, 1);
           While Name[Length(Name)]=' ' do
             Delete (Name,Length(Name),1);
                                                  (* Datei zum Lesen oeffnen *)
           Assign (f,Name);
           Reset (f);
                                               (* Groesse in kByte ermitteln *)
           RecSize := FileSize (f);
           kSize   := RecSize div 8;
           If RecSize mod 8 <> 0 then
             kSize := succ (kSize);
           Close (f)
           End
   End;

   (*------------------------------------------------------------------------*)

   Procedure SortDirectory (Var DirList :DirArray; NoEntries :Integer);

   Var ready    :Boolean;
       HelpStrg :DirStrg;

   Begin
     Repeat
       ready := true;
       For i:=1 to NoEntries-1 do
         If DirList[i].FileName > DirList[i+1].FileName then
           Begin
             HelpStrg              := DirList[i].FileName;
             DirList[i].FileName   := DirList[i+1].FileName;
             DirList[i+1].FileName := HelpStrg;
             ready := false;
           End;
     until ready;
   End;

   (*------------------------------------------------------------------------*)

   Procedure PrintDirectory (Device :DirOptions);

   Var Medium :Text;
       i      :Integer;

   Begin
     Case Device of                                (* Ausgabedatei eroeffnen *)
       CON: Assign (Medium,'CON:');
       LST: Assign (Medium,'LST:')
     end;
     WriteLn (Medium);
     WriteLn (Medium, 'Directory for User ', DirUser);
     WriteLn (Medium);
     If NoEntries<>0 then
       If not (Size in Options) then
         For i:=1 to NoEntries do
           Begin
             Write (Medium,DirList[i].FileName);
             If (i mod 5 = 0) or (i=NoEntries) then
               WriteLn (Medium)
             else
               Write (Medium,' : ');
           End
       else
         Begin
           WriteLn (Medium,' File              Recs      Bytes');
           WriteLn (Medium,'-----------------------------------');
           WriteLn (Medium);
           For i:=1 to NoEntries do
             With DirList[i] do
               WriteLn (Medium,' ',FileName,RecSize:10,kSize:10,'k')
         End
     else
       WriteLn (Medium,'No File');
   End;

(*---------------------------------------------------------------------------*)

Begin
  ParseFname (Mask, Drive, DirMask);
  SetDMA (DMA);
  p := SearchFirst (Mask);
  i := 0;
  While p<>255 do                      (* p=$FF: keine Datei mehr vorhanden *)
    Begin
      If DMA[p].User = DirUser then
        Begin
          i := succ (i);
          DirList[i].FileName := DMA[p].FileName;
          Insert (' ',DirList[i].FileName,9);
        End;
      p := SearchNext;
     End;
  NoEntries := i;
  If Sort in Options then
    SortDirectory (DirList, NoEntries);
  If Size in Options then
    ComputeFileSize (DirList, Drive, NoEntries);
  If CON  in Options then
    PrintDirectory (CON);
  If LST  in Options then
    PrintDirectory (LST)
End;

(*---------------------------------------------------------------------------*)
