{ Demo der Prozedur 'LeseSylk':
  x/y-Werte werden aus einer MP-Sylk-Datei in das REAL-Feld 'werte' gelesen
  und anschliessend ausgegeben.                                               }

PROGRAM lesemp;

CONST  abzeile = 3;      { Zeile, ab der aus Rechenblatt gelesen werden soll. }
      biszeile = 300;    { letzte, zulesende Zeile, falls soviel vorhanden.   }
      abspalte = 1;      { dito, jedoch fuer die Spalten.                     }
     bisspalte = 2;

TYPE  { REAL-Feld entspr. der obigen Dimensionen:
          1. Index: Index der Wertepaare,  2. Index: =1: x-Wert, =2: y-Wert.
        Sollen nicht x-/y-Wertepaare gelesen werden, sondern z.B. eine
        Matrix [1..n,1..m], muessen die Deklarationen der Konstanten und des
        Types 'rar' entspr. geaendert werden.                                 }

     rar = ARRAY [1..biszeile,1..bisspalte] OF REAL;

VAR  werte: rar;
       i,j: INTEGER;                { fuer Demo-Ausgabe der eingelesen Werte. }
      zmax,                       { bekommen von 'LeseSylk' die tatsaechliche }
      smax: INTEGER;              { Anzahl eingelesener Zeilen und Spalten.   }


{-----------------------------------------------------------------------------}
{ Hier wird das eigentliche Lesen der Daten aus einem MP-Rechenblatt er-
  ledigt. In dem REAL-Feld 'werte' werden die Werte ab dem Index 1 einer
  Dimension gespeichert. Die Ergebnis-Parameter 'zmax' und 'smax' geben
  die tatsaechliche Anzahl der gelesenen Zeilen und Spalten als oberen Index
  von 'werte' zurueck.                                                        }

PROCEDURE LeseSylk (VAR werte: rar;   { REAL-Array fuer eingelesene MP-Daten. }
                         ymin,        { Eingrenzung der zulesenden Werte:     }
                         xmin,        { linke, obere und rechte, untere Ecke  }
                         ymax,        { des Rechenblatts.                     }
                         xmax: INTEGER;
                    VAR  zmax,           { Rueckgabe der wirklich gelesenen   }
                         smax: INTEGER); { Zeilen und Spalten.                }


TYPE  SylkSatz = STRING [80];   { Sylk-Datensatz ist eine einfache Textzeile. }

VAR SylkDatei: TEXT;                    { entspr. besteht die Datei aus Text! }
         Satz: SylkSatz;
        FName: STRING [12];
         y, x,                                { y- u. x-Koord. im Rechenblatt }
         Spos: INTEGER;         { Position im Sylk-Satz waehrend Bearbeitung. }

  {---------------------------------------------------------------------------}
  PROCEDURE OeffneSylk;

  BEGIN
    WriteLn;
    Write('einzulesende SYLK-Datei angeben: ');
    ReadLn(FName);
    WriteLn;
    Assign(SylkDatei,FName);
    Reset(SylkDatei);
  END;

  {---------------------------------------------------------------------------}
  PROCEDURE LeseSatz;  { einen Sylk-Datensatz lesen, hier wg. String einfach. }

  BEGIN
    ReadLn(SylkDatei, Satz);
  END;

  {---------------------------------------------------------------------------}
  FUNCTION SatzTyp: CHAR;

  BEGIN
    SatzTyp := '?';                                    { unbekannter Satztyp. }
    Spos := Succ(Pos(';',Satz));
    IF Length(Satz) > 0 THEN
    BEGIN
      IF Spos = 3 THEN                         { ein Zeichen als Satzkennung, }
        SatzTyp := Satz[1]                     { wird spaeter behandelt.      }
      ELSE IF (Spos = 1) AND (Satz[1] = 'E') THEN      { Ende der Sylk-Datei. }
        SatzTyp := 'E'
      ELSE IF Spos = 4 THEN
        IF (Satz[1] = 'I') AND (Satz[2] = 'D') THEN      { ID der Sylk-Datei. }
          SatzTyp := 'I';
    END;
  END;

  {---------------------------------------------------------------------------}
  FUNCTION FeldTyp: CHAR;

  BEGIN
    FeldTyp := '?';                                    { unbekannter Feldtyp. }
    IF Spos < Length(Satz) THEN
      IF Satz[Spos] IN ['Y','X','K'] THEN    { z.Z. sind nur die interessant! }
      BEGIN
        FeldTyp := Satz[Spos];
        Spos := Succ(Spos);
      END;
  END;

  {---------------------------------------------------------------------------}
  FUNCTION Wert: REAL;        { numerischen Wert eines Sylk-Feldes ermitteln. }

  VAR     i: INTEGER;
       rvar: REAL;
        tmp: SylkSatz;

  BEGIN
    tmp := Copy(Satz, Spos, Length(Satz));
    i := 1;
    WHILE (tmp[i] <> ';') AND (i <= Length(tmp)) DO
    BEGIN                                      { Ende des num. Feldes suchen. }
      Spos := Succ(Spos);
      i := Succ(i);
    END;
    Spos := Succ(Spos);
    tmp := Copy(tmp, 1, Pred(i));
    Val(tmp, rvar, i);                     { Feldinhalt in num. Wert wandeln. }
    IF i <> 0 THEN                              { Umwandlung ist missglueckt! }
      rvar := 0;
    Wert := rvar;
  END;

  {---------------------------------------------------------------------------}
  { Wert eines num. Feldes aus dem Rechenblatt in das REAL-Feld uebernehmen,
    wenn innerhalb der spez. Indexgrenzen.                                    }

  PROCEDURE NimmWert;

  VAR ytmp, xtmp: INTEGER;

  BEGIN
    IF Satz[Spos] = '"' THEN                                     { Textfeld ! }
      Spos := Succ(Length(Satz))
    ELSE IF (y >= ymin) AND (y <= ymax) THEN
      IF (x >= xmin) AND (x <= xmax) THEN
        BEGIN
          xtmp := x-Pred(xmin);    { Anpassen des Blatt-Indexes an REAL-Feld. }
          ytmp := y-Pred(ymin);
          werte[ytmp,xtmp] := Wert;
          IF xtmp > smax THEN smax := xtmp;      { neue obere Index-Grenzen ? }
          IF ytmp > zmax THEN zmax := ytmp;
        END;
  END;

{-----------------------------------------------------------------------------}
BEGIN { LeseSylk }
  zmax := 0;
  smax := 0;
  FOR x := 1 TO xmax DO                              { REAL-Feld vorbesetzen. }
    FOR y := 1 TO ymax DO
      werte[y,x] := 0.0;
  OeffneSylk;
  LeseSatz;
  IF SatzTyp = 'I' THEN                                        { Sylk-Datei ? }
    WHILE NOT(Eof(SylkDatei)) DO
    BEGIN
      LeseSatz;
      CASE SatzTyp OF
        'C': WHILE Spos <= Length(Satz) DO    { interessanter Sylk-Datensatz! }
               CASE FeldTyp OF
                 'Y': y := Trunc(Wert);                { neuer Zeilen-Index ? }
                 'X': x := Trunc(Wert);               { neuer Spalten-Index ? }
                 'K': NimmWert;                                 { num. Feld ? }
               ELSE
                 BEGIN
                   WriteLn('*** Unbekanntes Feld:');
                   WriteLn(Satz);
                   WriteLn('^':Spos);
                   While (Satz[Spos] <> ';') AND (Spos <= Length(Satz)) DO
                     Spos := Succ(Spos);               { und zum naechsten.   }
                 END;
               END;
        'E': ;
      ELSE
        BEGIN
          WriteLn('*** Unbekannter Satz:');
          WriteLn(Satz);
        END;
      END;
    END
  ELSE
    WriteLn('*** ',FName,' ist keine MP-Sylk-Datei ! ***');
  Close(SylkDatei);
END;

BEGIN { lesemp }
  LeseSylk(werte,abzeile,abspalte,biszeile,bisspalte,zmax,smax);
  WriteLn('Die gelesenen Werte:');
  FOR i := 1 TO zmax DO
  BEGIN
    FOR j:=1 TO smax DO
      Write (werte[i,j]:10:3);
    Writeln;
  END;
  Write('Taste druecken...');
  Repeat Until KeyPressed;
END.
