(*****************************************************************************)
(*                                                                           *)
(*                          ---  SOLVE.INC  ---                              *)
(*                                                                           *)
(*          Unterprogramm-Modul zum Loesen einer Gleichung  g(x)=0           *)
(*                                                                           *)
(*****************************************************************************)


Function Solve (    a,b,                       (* Suchintervall          *)
                    d       :Real;             (* Suchschrittweite       *)
                    k       :Integer;          (* k-te Ableitung         *)
                Var Success :Boolean):Real;    (* Loesung gefunden?      *)


   Var  x,y,z,Lastx,Lasty  :Real;

   Begin
   Success := false;
   x := a;
   Lasty := fn (x,k);
   Lastx := x;
   x := x + d;
   If Lasty<>0 then
      Repeat
         y := fn (x,k);
         If y*Lasty <= 0 then                  (* Nullstelle lokalisiert *)
            Begin
            Success := true;
            If y<>0 then
               Repeat                   (* Anwendung der Regula falsi... *)
                  z := x;
                  x := x - y*(x-Lastx)/(y-Lasty);
                  Lastx := z;
                  Lasty := y;
                  y := fn (x,k)
               until abs(x-Lastx) < eps  (* ... bis Genauigkeit erreicht *)
            End
         else
            Begin
            Lasty := y;
            Lastx := x;
            x := x + d
            End
      until (x>b) or Success         (* suche bis Nullstelle gefunden... *)
   else                              (* ... oder Intervallende erreicht  *)
      Begin
      Success := true;
      x := x - d
      End;
   Solve := x
   End;
